#!/bin/sh
# $Id: t-mts-acc-8.sh,v 1.4 2007/02/17 04:49:04 ca Exp $
#
# ----------------------------------------
# test access map restrictions: EHLO
# uses smtpcnf.sh.
# ----------------------------------------
#
test -s stop && exit 1
#

SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../chkmts
fi
. ${SD}/common.sh

SCRIPTS=`cd ${SD} && pwd`

if ../smtps/smtps -VVV | ${GREP} SS_EHLO_ACCESS_CHK >/dev/null
then
  :
else
  echo "SKIPPED: $0: smtps not compiled with SS_EHLO_ACCESS_CHK"
  exit 0
fi

if test X"`uname`" != "XOSF1"
then
while getopts s:V FLAG
do
  case "${FLAG}" in
    s) SKIP="${SKIP} ${OPTARG}";;
    V) VERBOSE=true;;
  esac
done
shift `expr ${OPTIND} - 1`
fi

# cd to right directory (where the programs will be executed).
cd ../qmgr || exit 1

# create restrictions
# note: this only works if the client address is actually 127.0.0.1
# which is not the case when using FreeBSD jails.
${CAT} > ${ACC} <<EOF
ehlo:http://www.example.com/	error:550 5.7.0 no http://www.example.com/
EOF
# create map
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}
rm -f ${MTMAP}

${CAT} > ${SMXCNF} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4; }
smar { Log_Level = 16; nameserver = ${NS}; }
smtpc { Log_Level = 12; remote_port=${SNKPORT}; wait_for_server=4; }

smtps {
Log_Level = 12;
flags = {access, check_ehlo};
daemon_address = localhost:${SRVPORT};
relay_from = "No.host.";
relay_to = "No.host.";
wait_for_server=4;
}
EOF

# create aliases
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
user:	local:
EOF
# create map
rm -f ${ALIMAP}
${MM} < ${ALI}

${CAT} > ${MT} <<EOF
local.host	lmtp:
other.host	[127.0.0.1]
meta1.org	[127.0.0.1]
EOF

# run some test programs

# ----------------------------------------
# relaying allowed (client address) but not necessary
TEST=1
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -3 lmtpsock -Z 'smtpc2' -c '-E -e http://www.example.com/' -Ssender${TEST}@meta1.org -Ruser@local.host -E 0 -X 0 -A 0
then
  if ${GREP} 'EHLO=error, text=550 5.7.0 no http:' ${CLTL} >/dev/null
  then
    :
  else
    echo "$0: test ${TEST} failed: EHLO error not found"
    ERR=true
  fi
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
else
   ${VERBOSE} && echo "PASS: ${TEST}"
fi
fi


# ----------------------------------------
rm -f ${ACCMAP} ${ALIMAP} ${MTMAP}

# ----------------------------------------
# end
if test "${ERRS}" = "0"
then
  exit 0
else
  echo "${ERRS} error(s)"
  exit 1
fi
