/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-args2argv.c,v 1.2 2008/10/21 02:19:37 ca Exp $")
#include "sm/error.h"
#include "sm/assert.h"
#include "sm/ctype.h"
#include "sm/io.h"
#include "sm/str.h"
#include "util.h"
#include "sm/test.h"

#define SMAXLEN	256

static void
test_harness(void)
{
	sm_str_P str1;
	sm_ret_T ret;
	uint nelem;
	char **argv;

	str1 = sm_str_scpy0(NULL, "field1=value1 f2=v2", SMAXLEN);
	SM_TEST_E(str1 != NULL);

	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);
	SM_TEST(nelem == 2);
	SM_TEST_E(argv[0] != NULL);
	SM_TEST_E(argv[1] != NULL);
	SM_TEST_E(argv[2] == NULL);
	SM_TEST_E(strcmp(argv[0], "field1=value1") == 0);
	SM_TEST_E(strcmp(argv[1], "f2=v2") == 0);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f1= v1");
	SM_TEST_E(ret == SM_SUCCESS);
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f1 =v1");
	SM_TEST_E(ret == SM_SUCCESS);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f1=");
	SM_TEST_E(ret == SM_SUCCESS);
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "=v1");
	SM_TEST_E(ret == SM_SUCCESS);
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "=f1=");
	SM_TEST_E(ret == SM_SUCCESS);
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f1=v1    ");
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);
	SM_TEST_E(argv[0] != NULL);
	SM_TEST_E(argv[1] == NULL);
	SM_TEST_E(strcmp(argv[0], "f1=v1") == 0);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f1-/2=v1\\//!@#$%^&*()_-+=|~`[]{};:\"'<>,./?");
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);
	SM_TEST_E(argv[0] != NULL);
	SM_TEST_E(argv[1] == NULL);
	SM_TEST_E(strcmp(argv[0], "f1-/2=v1\\//!@#$%^&*()_-+=|~`[]{};:\"'<>,./?") == 0);

	sm_str_clr(str1);
	ret = sm_str_scat0(str1, "f2=v2 novalue");
	SM_TEST(ret == SM_SUCCESS);
	ret = args2argv(str1, &nelem, &argv);
	SM_TEST(SM_SUCCESS == ret);
	SM_TEST_E(argv[0] != NULL);
	SM_TEST_E(argv[1] != NULL);
	SM_TEST_E(argv[2] == NULL);
	SM_TEST_E(strcmp(argv[0], "f2=v2") == 0);
	SM_TEST_E(strcmp(argv[1], "novalue") == 0);


	SM_STR_FREE(str1);

 error:
	return;
}

int
main(int argc, char *argv[])
{
	sm_test_begin(argc, argv, "test args2argv()");
	test_harness();
	return sm_test_end();
}
