/*
 * Copyright (c) 2004, 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-seq-1.c,v 1.14 2009/06/17 05:03:50 ca Exp $")

#include "sm/error.h"
#include "sm/heap.h"
#include "sm/memops.h"
#include "sm/test.h"
#include "sm/maps.h"
#include "sm/mapc.h"
#include "sm/map.h"
#include "sm/mapclasses.h"
#include "sm/bdb.h"

#include "sm/io.h"

int Verbose = 0;

#include "t-seq-common.c"

static bool
testh(bool usename, bool usebht)
{
	sm_ret_T ret;
	sm_maps_P maps;
	sm_map_P map1, map2, map3;
	sm_cstr_P mname1, mname2;
	sm_cstr_P mtype, mname;
	sm_str_P lhs, rhs;
	sm_str_T key, data;

	maps = NULL;
	mtype = mname = NULL;
	lhs = rhs = NULL;
	mname1 = mname2 = NULL;

	ret = sm_maps_init(&maps);
	SM_TEST(maps != NULL);
	if (maps == NULL)
		return false;
	SM_TEST(sm_is_success(ret));

	lhs = sm_str_new(NULL, 256, 1024);
	SM_TEST_E(lhs != NULL);
	rhs = sm_str_new(NULL, 256, 1024);
	SM_TEST_E(rhs != NULL);
	ret = sm_bht_class_create(maps);
	SM_TEST_E(sm_is_success(ret));
	ret = sm_seq_class_create(maps);
	SM_TEST_E(sm_is_success(ret));

	ret = crt_map(maps, &map1, MAPC_NAME1, MAPC_TYPE1);
	SM_TEST_E(sm_is_success(ret));
	mname1 = sm_cstr_scpyn0((const uchar *)MAPC_NAME1, strlen(MAPC_NAME1));
	ret = sm_error_temp(SM_EM_MAP, ENOMEM);
	SM_TEST_E(mname1 != NULL);

#define MAPC_TYPEBHT     "bhtable"
#define MAPC_NAMEBHT     "bht2"

	if (usebht) {
		ret = crt_map(maps, &map2, MAPC_NAMEBHT, MAPC_TYPEBHT);
		SM_TEST_E(sm_is_success(ret));
		mname2 = sm_cstr_scpyn0((const uchar *)MAPC_NAMEBHT, strlen(MAPC_NAMEBHT));
	} else {
		ret = crt_bdb_map(maps, &map2, MAPC_NAME2, MAPC_TYPE2, MAPC_FILE2);
		SM_TEST_E(sm_is_success(ret));
		mname2 = sm_cstr_scpyn0((const uchar *)MAPC_NAME2, strlen(MAPC_NAME2));
	}
	ret = sm_error_temp(SM_EM_MAP, ENOMEM);
	SM_TEST_E(mname2 != NULL);

	/* populate BTH map */
	SM_POP_MAP1;

	/* populate BDB map */
	SM_POP_MAP2;

	/* perform some operations ... */
	SM_LOOKUP_TEST(map1, KEY1_1, KEY1_2, KEY1_3);

	/* perform some operations ... */
	SM_LOOKUP_TEST(map2, KEY2_1, KEY2_2, KEY2_3);

	SM_REWRITE_TEST(map1, KEY1_1, DATA1_1);
	SM_REWRITE_TEST(map1, KEY1_2, DATA1_2);

	/* create sequence map */
	mtype = mname = NULL;
	mtype = sm_cstr_scpyn0((const uchar *)MAPC_TYPE3, strlen(MAPC_TYPE3));
	ret = sm_error_temp(SM_EM_MAP, ENOMEM);
	SM_TEST_E(mtype != NULL);
	mname = sm_cstr_scpyn0((const uchar *)MAPC_NAME3, strlen(MAPC_NAME3));
	SM_TEST_E(mname != NULL);
	map3 = NULL;
	ret = sm_map_create(maps, mtype, 0, &map3);
	SM_TEST_E(sm_is_success(ret));
	if (usename) {
		ret = sm_map_setopt(map3,
				SMPO_MAPNAME, mname1,
				SMPO_MAPNAME, mname2,
				SMPO_END);
	}
	else {
		ret = sm_map_setopt(map3,
				SMPO_MAP, map1,
				SMPO_MAP, map2,
				SMPO_END);
	}
	SM_TEST_E(sm_is_success(ret));
	ret = sm_map_open(maps, mname, mtype, 0, NULL, SMAP_MODE_RDWR, &map3,
			SMPO_END);
	SM_TEST_E(sm_is_success(ret));
	SM_CSTR_FREE(mtype);
	SM_CSTR_FREE(mname);

	/* perform some operations ... */
	SM_LOOKUP_TEST(map3, KEY1_1, KEY1_2, KEY1_3);
	SM_LOOKUP_TEST(map3, KEY2_1, KEY2_2, KEY2_3);
	SM_REWRITE_TEST(map3, KEY1_1, DATA1_1);
	SM_REWRITE_TEST(map3, KEY1_2, DATA1_2);

	SM_CLOSE_MAPS;

	ret = sm_maps_term(maps);
	SM_TEST(sm_is_success(ret));
	SM_STR_FREE(lhs);
	SM_STR_FREE(rhs);
	return true;

  error:
	sm_maps_term(maps);
	return false;
}


int
main(int argc, char *argv[])
{
	int c;

	while ((c = getopt(argc, argv, "H:s:V")) != -1)
	{
		switch (c)
		{
		  default:
			return 1;
		}
	}
	sm_test_begin(argc, argv, "test sequence map 1");
	if (!testh(false, false)) goto end;
	if (!testh(false, true)) goto end;
	if (!testh(true, false)) goto end;
	if (!testh(true, true)) goto end;

  end:
	return sm_test_end();
}
